
class
	CUBE_SOLVER
	
inherit
	EXCEPTIONS
	
feature find_solution(cube: IDEAL_CUBE): ARRAYED_LIST[PLAYER_MOVE] is
		require
			cube /= Void
		local	
			i: INTEGER
			solution_system_array: ARRAY_LIST
			moves_line_system_string: SYSTEM_STRING
			moves_array: ARRAYED_LIST[STRING]
			move_string: STRING
		do
			create Result.make(0)
			load_rubik_solver_dll
			solution_system_array := call_solver(cube_colors_for_solver(cube))
			if solution_system_array = Void then
				raise("Cube solver didn't return anything.")
			end
			from i := 0 until i >= solution_system_array.count loop
				moves_line_system_string ?= solution_system_array.item(i)
				if moves_line_system_string = Void then
					raise("Solution contains not a string object.")
				end
				moves_array := parse_moves_string(moves_line_system_string)
				from moves_array.start until moves_array.off loop
					move_string := moves_array.item
					if not valid_move(move_string) then
						raise("Wrong move from solver: " + move_string)
					end
					Result.append(string_to_moves(move_string))
					moves_array.forth
				end
				i := i + 1
			end
			Result := remove_redundant_moves(Result)
			
			if not valid_solution(cube, Result) then
				raise("Solver failed to solve cube.")
			end
		ensure
			Result /= Void 
			not Result.has(Void) 
			valid_solution(cube, Result)
		end
	
feature	valid_solution(cube: IDEAL_CUBE solution: ARRAYED_LIST[PLAYER_MOVE]): BOOLEAN is
		require
			cube /= Void
			solution /= Void and then not solution.has(Void)
		local
			cube_copy: like cube
		do
			cube_copy := deep_clone(cube)
			solution.do_all(agent cube_copy.play_move)
			Result := cube_copy.is_solved 
		end
	
feature valid_move(s: STRING): BOOLEAN is
		require
			s /= Void
		do
			Result := valid_short_move(s) or valid_long_move(s) 
		end
	
feature valid_long_move(s: STRING): BOOLEAN is
		require
			s /= Void
		do
			Result := s.count = 3 and then (("-+").has(s.item(1)) and s.item(2) = 'r' and ("rfu").has(s.item(3)))
		end
	
feature valid_short_move(s: STRING): BOOLEAN is
		require
			s /= Void
		do
			Result := s.count = 2 and then (("-+").has(s.item(1)) and ("urlfdb").has(s.item(2)))
		end
	
feature {NONE} -- access to RUBIK_SOLVER
	
	call_solver(colors: ARRAY_LIST): ARRAY_LIST is
		require
			colors /= Void
		do
			Result := feature {RUBIK_SOLVER}.solve(colors)
		end	
	
	load_rubik_solver_dll is
		local 
			exception_occured: BOOLEAN
			dummy: ASSEMBLY
			solver_assembly_name: ASSEMBLY_NAME
		do
			if not exception_occured then
				create solver_assembly_name.make
				solver_assembly_name.set_name("cube solver")
				dummy := feature {ASSEMBLY}.load(solver_assembly_name)
			else
				raise("Can't load cube solver.dll or amzi.dll")
			end
		rescue
			if not exception_occured then
				exception_occured := True
				retry
			end
		end
	
feature {NONE} -- Removing redundant moves
	
	remove_redundant_moves(original: ARRAYED_LIST[PLAYER_MOVE]): ARRAYED_LIST[PLAYER_MOVE] is
		require
			original /= Void and then not original.has(Void)
		local 
			old_count: INTEGER
			no_shrink: BOOLEAN
		do
			Result := clone(original)
			from until no_shrink loop 
				old_count := Result.count
				combine_similar_moves(Result)
				no_shrink := old_count = Result.count
			end 
		ensure
			Result /= Void 
			Result.count <= original.count 
			not Result.has(Void)
		end
	
	combine_similar_moves(data: ARRAYED_LIST[PLAYER_MOVE]) is
		require
			data /= Void and then not data.has(Void)
		local 
			i, next_similar: INTEGER
			move: PLAYER_MOVE
		do
			from i := 1 until i > data.count loop
				next_similar := find_similar(data.i_th(i), i + 1, data)
				if next_similar > 0 then
					if data.i_th(i).rotate_count.count + data.i_th(next_similar).rotate_count.count \\ 4 = 0 then
						data.go_i_th(next_similar)
						data.remove
						data.go_i_th(i)
						data.remove
					else	
						create move.make(data.i_th(i).plane, data.i_th(i).rotate_count + data.i_th(next_similar).rotate_count)
						data.go_i_th(i)
						data.put(move)
						data.go_i_th(next_similar)
						data.remove
					end
				else
					i := i + 1	
				end
			end
		ensure
			data.count <= old data.count
			not data.has(Void)
		end
	
	find_similar(move: PLAYER_MOVE start_index: INTEGER moves: ARRAYED_LIST[PLAYER_MOVE]): INTEGER is
		require
			move /= Void
			start_index > 0
			moves /= Void and then not moves.has(Void)
		local 
			i: INTEGER
			finished: BOOLEAN
		do
			from i := start_index until i > moves.count or finished loop
				if not moves.i_th(i).plane.orientation.is_equal(move.plane.orientation) then
					finished := True
				else
					if moves.i_th(i).plane.index.is_equal(move.plane.index) then
						Result := i
						finished := True
					end
				end
				i := i + 1
			end
		ensure
			Result = 0 or (Result >= start_index and Result <= moves.count)
		end
	
feature {NONE} -- Solution decoding
	
	parse_moves_string(s: STRING): ARRAYED_LIST[STRING] is
		require
			s /= Void
		do
			s.prune_all(' ')
			s.prune_all('[')
			s.prune_all(']')
			create Result.make(0)
			Result.append(s.split(','))
		ensure
			Result /= Void and then not Result.has(Void)	
		end
	
	string_to_moves(s: STRING): ARRAYED_LIST[PLAYER_MOVE] is
		require
			valid_move(s)
		do
			if valid_short_move(s) then
				create Result.make(1)
				Result.extend(short_string_to_move(s))
			else
				Result := long_string_to_moves(s)
			end	
		ensure
			Result /= Void
			Result.count = 1 or	Result.count = 3
			not Result.has(Void)
		end
	
	long_string_to_moves(s: STRING): ARRAYED_LIST[PLAYER_MOVE] is
		require
			valid_long_move(s)
		do
			if s.is_equal("-rr") then
				Result := rotate_whole_cube(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {ROTATE_COUNT}.make(1))
			elseif s.is_equal("+rr") then	
				Result := rotate_whole_cube(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {ROTATE_COUNT}.make(-1))
			elseif s.is_equal("-rf") then	
				Result := rotate_whole_cube(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {ROTATE_COUNT}.make(-1))
			elseif s.is_equal("+rf") then	
				Result := rotate_whole_cube(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {ROTATE_COUNT}.make(1))	
			elseif s.is_equal("-ru") then	
				Result := rotate_whole_cube(feature {ORIENTATION}.Horizontal, create {ROTATE_COUNT}.make(-1))	
			elseif s.is_equal("+ru") then	
				Result := rotate_whole_cube(feature {ORIENTATION}.Horizontal, create {ROTATE_COUNT}.make(1))		
			end
		ensure
			Result /= Void 
			Result.count = 3	
			not Result.has(Void)
		end
	
	short_string_to_move(s: STRING): PLAYER_MOVE is
		require
			valid_short_move(s)
		local
			p: CUBE_PLANE_PARAMETERS	
		do
			if s.is_equal("-u") then
				create p.make(feature {ORIENTATION}.Horizontal, create {IDEAL_CUBE_INDEX}.make(1))			
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			elseif s.is_equal("+u") then
				create p.make(feature {ORIENTATION}.Horizontal, create {IDEAL_CUBE_INDEX}.make(1))			
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("-r") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {IDEAL_CUBE_INDEX}.make(-1)) 
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("+r") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {IDEAL_CUBE_INDEX}.make(-1)) 			
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			elseif s.is_equal("-l") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {IDEAL_CUBE_INDEX}.make(1))
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			elseif s.is_equal("+l") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_y, create {IDEAL_CUBE_INDEX}.make(1))
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("-f") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {IDEAL_CUBE_INDEX}.make(1))
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			elseif s.is_equal("+f") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {IDEAL_CUBE_INDEX}.make(1))
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("-d") then
				create p.make(feature {ORIENTATION}.Horizontal, create {IDEAL_CUBE_INDEX}.make(-1))
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("+d") then
				create p.make(feature {ORIENTATION}.Horizontal, create {IDEAL_CUBE_INDEX}.make(-1))
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			elseif s.is_equal("-b") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {IDEAL_CUBE_INDEX}.make(-1))
				create Result.make(p, create {ROTATE_COUNT}.make(1))			
			elseif s.is_equal("+b") then
				create p.make(feature {ORIENTATION}.Vertical_perpendicular_to_x, create {IDEAL_CUBE_INDEX}.make(-1))			
				create Result.make(p, create {ROTATE_COUNT}.make(-1))			
			end
		ensure
			Result /= Void	
		end
	
	rotate_whole_cube(orientation: INTEGER rotate_count: ROTATE_COUNT): ARRAYED_LIST[PLAYER_MOVE] is
		require
			rotate_count /= Void
		local
			i: INTEGER
			move: PLAYER_MOVE
		do
			create Result.make(3)
			from i := -1 until i > 1 loop
				create move.make(create {CUBE_PLANE_PARAMETERS}.make(orientation, create {IDEAL_CUBE_INDEX}.make(i)), rotate_count)			
				Result.extend(move)
				i := i + 1
			end
		ensure
			Result /= Void 
			Result.count = 3
			not Result.has(Void)
		end
	
feature {NONE} -- Encoding cube position
	
	cube_colors_for_solver(cube: IDEAL_CUBE): ARRAY_LIST is
		require
			cube /= Void
		local
			dummy: INTEGER
			x, y, z: INTEGER
		do
			create Result.make
			
			x := -1
			from z := -1 until z > 1 loop
				from y := 1 until y < -1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Back))
					y := y - 1
				end
				z := z + 1
			end	
			
			y := 1
			from x := -1 until x > 1 loop
				from z := -1 until z > 1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Left))
					z := z + 1
				end
				x := x + 1
			end	
			
			z := 1
			from x := -1 until x > 1 loop
				from y := 1 until y < -1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Top))
					y := y - 1
				end
				x := x + 1
			end	
			
			y := -1
			from x := -1 until x > 1 loop
				from z := 1 until z < -1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Right))
					z := z - 1
				end
				x := x + 1
			end	
			
			x := 1
			from z := 1 until z < -1 loop
				from y := 1 until y < -1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Front))
					y := y - 1
				end
				z := z - 1
			end	
			
			z := -1
			from x := 1 until x < -1 loop
				from y := 1 until y < -1 loop
					dummy := Result.add(cell_color(x, y, z, cube, Direction.Bottom))
					y := y - 1
				end
				x := x - 1
			end	
		ensure
			Result /= Void and then	Result.count = 54
		end
	
	cell_color(x, y, z: INTEGER cube: IDEAL_CUBE at_direction: DIRECTION): INTEGER is
		require
			x = -1 or x = 0 or x = 1
			y = -1 or y = 0 or y = 1
			z = -1 or z = 0 or z = 1
			cube /= Void
			at_direction /= Void
		local
			brick: IDEAL_BRICK
			position: IDEAL_BRICK_POSITION
		do
			create position.make(create {IDEAL_CUBE_INDEX}.make(x), 
								 create {IDEAL_CUBE_INDEX}.make(y), 
								 create {IDEAL_CUBE_INDEX}.make(z))
			brick := cube.select_brick(position)
			Result := color_to_int(brick.colors.select_color(at_direction))
		ensure
			Result >= 0 and Result < 6	
		end
	
	color_to_int(color: INTEGER): INTEGER is
		do
			inspect 
				color
			when feature {COLOR}.Pink then 
				Result := 3
			when feature {COLOR}.Red then 
				Result := 4
			when feature {COLOR}.Yellow then 
				Result := 2
			when feature {COLOR}.Blue then 
				Result := 1
			when feature {COLOR}.Green then 
				Result := 0
			when feature {COLOR}.White then 
				Result := 5
			end
		ensure
			Result >= 0 and Result < 6	
		end
	
feature	{NONE} Direction: DIRECTION is
		once
			create Result
		end           
	
end 
