indexing
	description: "Standard Direct3D functionality"
	
deferred class
    D3D_APP
	
inherit
	FULLSCREEN_VIEW
	WIREFRAME_VIEW
	MISC	
	EXCEPTIONS
	
feature {NONE} make(new_window_handle: POINTER new_form: WINFORMS_FORM 
		start_in_fullscreen, new_software_rendering: BOOLEAN new_background: DRAWING_COLOR
		start_size: DRAWING_SIZE) is
		require
			new_window_handle /= Void
			new_form /= Void
		do
			window_handle := new_window_handle
			form := new_form
			fullscreen := start_in_fullscreen
			software_rendering := new_software_rendering
			background := new_background
			create update_callback.make(Current, $not_thread_safe_update)		
			create fullscreen_toggle_callback.make(Current, $not_thread_safe_fullscreen_toggle)		
			initialize_graphics(start_size)
		end
	
feature -- Public features
	
	update is
		local 
			dummy: SYSTEM_OBJECT
		do
			if form.invoke_required then	
				dummy := form.invoke(update_callback)		
			else
				not_thread_safe_update
			end
		end	
	
	toggle_fullscreen_mode is
		local 
			dummy: SYSTEM_OBJECT
		do
			if form.invoke_required then	
				dummy := form.invoke(fullscreen_toggle_callback)		
			else
				not_thread_safe_fullscreen_toggle
			end
		end
	
	toggle_wireframe_mode is
		do
			wireframe := not wireframe
			render_device.render_state.set_fill_mode(current_fill_mode)
			update
		end		
	
feature {NONE} -- Deferred features 
	
	custom_render is
		deferred
		end
	
	custom_initialize_graphics is
		deferred
		end	
	
feature {NONE} -- Callbacks
	
	not_thread_safe_update is
		require
			not form.invoke_required 	
		do
			render
		end	
	
	not_thread_safe_fullscreen_toggle is
		require
			not form.invoke_required 	
		do
			fullscreen := not fullscreen
			if fullscreen then
				form.set_window_state(feature {WINFORMS_FORM_WINDOW_STATE}.normal) -- If window is maximized, it'll be hard to restore window mode later 
				window_size := form.size
				form.set_size(fullscreen_size)
				reset_device
			else
				reset_device
				form.set_size(window_size)
			end	
		end
	
feature {NONE} -- Direct3D
	
	initialize_graphics(start_size: DRAWING_SIZE) is 
		do
			if fullscreen then
				window_size := start_size
				form.set_size(fullscreen_size)
			else
				form.set_size(start_size)	
			end		
			try_to_create_device
			custom_initialize_graphics
		end
	
	try_to_create_device is
		local
			exception_occured: BOOLEAN
		do
			if not exception_occured then
				create render_device.make(0, 
										  rendering_mode, 
										  window_handle, 
										  feature {CREATE_FLAGS}.software_vertex_processing | 
										  feature {CREATE_FLAGS}.fpu_preserve | 
										  feature {CREATE_FLAGS}.multi_threaded, 
										  present_parameters)
				render_device.add_device_reset(create {EVENT_HANDLER}.make(Current, $on_reset_device))  
			else
				raise("Your video system doesn't suport Direct3D rendering in current mode.")
			end
		rescue
			if not exception_occured then
				exception_occured := True
				retry
			end
		end
	
	render is 
		local
			b: DRAWING_COLOR
		do
			if device_lost then
				try_to_restore_lost_device
			end
			if not device_lost then
				b := background
				render_device.clear(feature {CLEAR_FLAGS}.Target | feature {CLEAR_FLAGS}.ZBuffer, b.to_argb, 1.0, 0)
				render_device.begin_scene
				
				custom_render
				
				render_device.end_scene
				try_to_present
			end
		end
	
	try_to_restore_lost_device is
		local
			status: INTEGER
		do
			if device_lost then 
				feature {SYSTEM_THREAD}.sleep(100)
				if render_device.check_cooperative_level_integer($status) then
					device_lost := False
				elseif status = feature {RESULT_CODE}.device_not_reset.to_integer then
					device_lost := False
					reset_device
				end	
			end
		end
	
	try_to_present is
		do
			if not device_lost then
				render_device.present
			end
		rescue
			if last_exception_type.is_equal("Microsoft.DirectX.Direct3D.DeviceLostException") then
				device_lost := True
				form.invalidate -- try to paint again, restoring lost device on the way
				retry
			end
		end
	
	reset_device is
		do
			if not device_lost then
				render_device.reset(present_parameters)
			end
		rescue
			if last_exception_type.is_equal("Microsoft.DirectX.Direct3D.DeviceLostException") then
				device_lost := True
				retry
			end	
		end		
	
	on_reset_device(sender: SYSTEM_OBJECT args: EVENT_ARGS) is
		do
			if not (render_device.render_state.fill_mode = current_fill_mode) then
				render_device.render_state.set_fill_mode(current_fill_mode)		
			end
		end
	
	
	
feature {NONE} -- Rendering settings
	
	rendering_mode: DEVICE_TYPE is
		do
			if software_rendering then
				Result := feature {DEVICE_TYPE}.reference_
			else
				Result := feature {DEVICE_TYPE}.hardware
			end
		end	
	
	current_fill_mode: FILL_MODE is
		do
			if wireframe then
				Result := feature {FILL_MODE}.wire_frame
			else
				Result := feature {FILL_MODE}.solid
			end	
		end	
	
	present_parameters: PRESENT_PARAMETERS is
		local
			current_mode: DISPLAY_MODE
		do
			create Result.make
			Result.set_windowed(not fullscreen)
			if fullscreen then
				current_mode := feature {MANAGER}.adapters.default.current_display_mode
				Result.set_back_buffer_width(current_mode.width)
				Result.set_back_buffer_height(current_mode.height)
				Result.set_full_screen_refresh_rate_in_hz(current_mode.refresh_rate)  
				Result.set_back_buffer_format(current_mode.format)
				Result.set_swap_effect(feature {SWAP_EFFECT}.Flip)
			else
				Result.set_swap_effect(feature {SWAP_EFFECT}.Discard)
			end				
			Result.set_back_buffer_count(1)
			Result.set_enable_auto_depth_stencil(True) 
			Result.set_auto_depth_stencil_format(feature {DEPTH_FORMAT}.D16) 
		ensure
			Result /= Void
		end	
	
	fullscreen_size: DRAWING_SIZE is
		local
			current_mode: DISPLAY_MODE
		do
			current_mode := feature {MANAGER}.adapters.default.current_display_mode
			Result.make(current_mode.width, current_mode.height)
		end	
	
feature {NONE}	
	render_device: DEVICE	
	window_handle: POINTER
	form: WINFORMS_FORM
	software_rendering: BOOLEAN
	window_size: DRAWING_SIZE
	background: DRAWING_COLOR
	device_lost: BOOLEAN
	update_callback, fullscreen_toggle_callback: WAIT_CALLBACK
	
invariant
	window_handle /= Void
	render_device /= Void
	update_callback /= Void
	fullscreen_toggle_callback /= Void
	form /= Void
	
end 
